/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 8 -*- */

/* commands.c -- Commands for musicdaemon player
 *
 * Copyright (C) 2002-2003 Petri Lahtinen <pjl@timefade.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


#include "commands.h"
#include "log.h"
#include "plugin.h"
#include "playlist.h"
#include "musicd.h"
#include "settings.h"


int cmd_play(core *core, client *user, char *param)
{
    song *song;
    int free_song = 0;
    
    if(param)
	log_printf(core, "cmd_play() called with param: %s",param);
    else
	log_printf(core, "cmd_play() called");

    if(!param)
    {
	if(has_flag(core->status, STATUS_RANDOM) && !core->rands)
	    cmd_next(core, user, NULL);
	
	set_flag(core->status, STATUS_PLAYLIST);
	if(list_size(&core->active_playlist->songs))
	{   
	    if(!core->active_playlist->position)
		core->active_playlist->position = 1;

	    song = list_get(core->active_playlist->position, 
			    &core->active_playlist->songs);
	    if(!song)
	    {   
		core->active_playlist->position = 1;
		if(!has_flag(core->status, STATUS_LOOP))
		    return(1);

		song = list_get(core->active_playlist->position, 
				&core->active_playlist->songs);
		if(!song)
		    return(1);
	    }
	}
	else
	    return(1);
    }
    else
    {
	unset_flag(core->status, STATUS_PLAYLIST);
	song = calloc(1, sizeof(struct song));
	free_song = 1;
	song->file = param;
    }

    if(!song->info.filename)
    {
	if(has_flag(core->status, STATUS_PLAYLIST))
	    update_songinfo(core, song, core->active_playlist->position);
	else
	    update_songinfo(core, song, -1);
    }

    if(play_song(core, song))
    {
	log_printf(core, "No plugin found for \"%s\".", param);
	if(free_song)
	{
	    free_metainfo(&song->info);
	    free(song);
	}
	return(1);
    }
    if(free_song)
    {
	free_metainfo(&song->info);
	free(song);
    }
    return(0);
}


int cmd_pause(core *core, client *user, char *param)
{
    if(param || user) {}
    log_printf(core, "cmd_pause() called");
    swap_flag(core->status, STATUS_PAUSED);
    (*core->active_output->pause)(has_flag(core->status, STATUS_PAUSED));
    if(has_flag(core->status, STATUS_PAUSED))
    {
	broadcast(core,"PAUSED\n");
	core->stats.playtime += time(0) - core->stats.song_start;
    }
    else
    {
	broadcast(core,"PLAY\n");
	core->stats.song_start = time(0);
    }
    
    return(0);
}

int cmd_stop(core *core, client *user, char *param)
{
    if(param || user) {}
    if(has_flag(core->status, STATUS_PLAYING))
    {
	core->stats.playtime += time(0) - core->stats.song_start;
	(*core->active_input->stop)();
	(*core->active_output->pause)(0);
	core->active_input = 0;
	unset_flag(core->status, STATUS_PLAYING);
	unset_flag(core->status, STATUS_PAUSED);
    }
    broadcast(core, "STOP\n");
    log_printf(core, "cmd_stop() called");
    return(0);
}

int cmd_seek(core *core, client *user, char *param)
{
    long int position;
    log_printf(core, "cmd_seek() called");
    if(!param)
    {
	send_to_client(core, user, "ERROR: Missing seek position.\n");
	return(-1);
    }
    
    position = strtol(param, NULL, 10);
    
    if(core->active_input->seek)
	(*core->active_input->seek)(position);
    
    return(0);
}

int cmd_next(core *core, client *user, char *param)
{
    if(param || user) {}
    log_printf(core, "cmd_next() called");
    cmd_stop(core, user, NULL);
    if(has_flag(core->status, STATUS_RANDOM))
    {
	double size = list_size(&core->active_playlist->songs);
	core->active_playlist->position = 
	    1+(int) ((size)*rand()/(RAND_MAX+1.0));
	core->rands++;
    }
    else
	core->active_playlist->position++;
    cmd_play(core, user, NULL);
    return(0);
}

int cmd_prev(core *core, client *user, char *param)
{
    int i;
    if(param || user) {}
    log_printf(core, "cmd_prev() called");
    cmd_stop(core, user, NULL);
    if(has_flag(core->status, STATUS_RANDOM))
    {
	double size = list_size(&core->active_playlist->songs);
	if(core->rands)
	    core->rands--;
	srand(core->seed);
	for(i = 1; i < core->rands; i++)
	    rand();
	core->active_playlist->position =
	    1+(int) ((size)*rand()/(RAND_MAX+1.0));
    }
    else
	core->active_playlist->position--;
    cmd_play(core, user, NULL);
    return(0);
}

int cmd_add(core *core, client *user, char *param)
{
    song *song;
    log_printf(core, "cmd_add() called");

    if(!param)
    {
	send_to_client(core, user, "ERROR: Missing filename.\n");
	return(-1);
    }
    song = calloc(1,sizeof(struct song));
    song->file = strdup(param);
    list_add(song, list_size(&core->active_playlist->songs) + 1, 
	     &core->active_playlist->songs);
    return(0);
}

int cmd_add_at(core *core, client *user, char *param)
{
    if(core || param || user) {}
    log_printf(core, "cmd_add_at() called");
    return(0);
}

int cmd_clear(core *core, client *user, char *param)
{
    playlist *list = core->active_playlist;
    if(param || user) {}
    log_printf(core, "cmd_clear() called");

    while(list->songs.start) 
    {
	free(((song *)list->songs.start->data)->file);
	free_metainfo(&((song *)list->songs.start->data)->info);
	free(list->songs.start->data);
	list_delete(0, &list->songs);
    }
    
    return(0);
}

int cmd_show(core *core, client *user, char *param)
{
    list_node *song;
    if(param) {}
    log_printf(core, "cmd_show() called");
    list_for_each(song, &core->active_playlist->songs)
	client_printf(core, user, "%s\n", ((struct song*)song->data)->file);
    return(0);
}
    
int cmd_jump(core *core, client *user, char *param)
{
    int position;
    if(core || param || user) {}
    log_printf(core, "cmd_jump() called");

    if(!param)
    {
	send_to_client(core, user, "ERROR: Missing postition.\n");
	return(-1);
    }

    position = atoi(param);
    if (position > list_size(&core->active_playlist->songs))
    {
	send_to_client(core, user, "ERROR: Invalid position.\n");
	return 1;
    }

    core->active_playlist->position = position;
    cmd_stop(core, user, NULL);
    cmd_play(core, user, NULL);
    return(0);
}

int cmd_search(core *core, client *user, char *param)
{
    if(core || param || user) {}
    log_printf(core, "cmd_search() called");
    return(0);
}

int cmd_load(core *core, client *user, char *param)
{
    char buf[MAX_MSG_SIZE+1];
    FILE *fd;
    song *song;
    
    if(core || param || user) {}
    log_printf(core, "cmd_load() called");
    if(!param)
    {   
	send_to_client(core, user, "ERROR: Missing filename.\n");
	return(-1);
    }
    fd = fopen(param, "r");
    if(!fd)
    {
	send_to_client(core, user, "ERROR: Unable to open file.\n");
	return(-1);
    }
    while(!feof(fd))
    {
	if(!fgets(buf, MAX_MSG_SIZE, fd))
	    break;
	if(!strchr(buf, '\n'))
	{   
	    while(!strchr(buf, '\n'))
	    {
		*buf = 0;
		fgets(buf, sizeof(buf), fd);
	    }
	    log_printf(core, "Too long line in playlist (skipped).\n");
	    continue;
	}
	if(buf[0] == '#' || buf[0] == 0)
	    continue;
	*(char*)strchr(buf, '\n') = 0;
	song = calloc(1,sizeof(struct song));
	song->file = strdup(buf);
	list_add(song, list_size(&core->active_playlist->songs) + 1,
		 &core->active_playlist->songs);
    }
    fclose(fd);
    return(0);
}

int cmd_remove(core *core, client *user, char *param)
{
    int position;
    if(core || param || user) {}
    log_printf(core, "cmd_remove() called");
    if(!param)
    {
	send_to_client(core, user, "ERROR: Missing postition.\n");
	return(-1);
    }
    position = atoi(param);
    if(list_delete(position, &core->active_playlist->songs) == -1)
    {
	send_to_client(core, user, "ERROR: Invalid position.\n");
	return(-1);
    }
    return(0);
}

int cmd_shutdown(core *core, client *user, char *param)
{
    if(core || param || user) {}
    if (core)
    {
	set_flag(core->status, STATUS_QUIT);
    }
    log_printf(core, "cmd_shutdown() called");
    return 0;
}

int cmd_set(core *core, client *user, char *param)
{
    char *variable = param;
    char *value;
    
    log_printf(core, "cmd_set() called");

    if(!variable)
    {
	settings_print_all(core, user);
	return(0);
    }
    
    value = strchr(param, '=');

    if(!value)
    {
	setting_print(core, user, variable);
	return(0);
    }
    
    *value = 0;
    value++;
    
    switch( set(core, variable, value) )
    {
        case -1:
	    send_to_client(core, user, "ERROR: Invalid value.\n");
	    break;

        case 0:
	    setting_print(core, user, variable);
	    break;
	    
        case 1:
	    send_to_client(core, user, "ERROR: No such variable.\n");
	    break;

        default:
	    log_printf(core, "set() returned unknown value");
	    send_to_client(core, user, 
		"BUG: Internal error: set() returned unknown value\n");
	    break;
    }
    
    return(0);
}

int cmd_move(core *core, client *user, char *param)
{
    if(core || param || user) {}
    log_printf(core, "cmd_move() called");
    return(0);
}

int cmd_stats(core *core, client *user, char *param)
{
    if(core || param || user) {}
    log_printf(core, "cmd_stats() called");
    client_printf(core, user, "UPTIME: %d\n", time(0) - core->stats.uptime);
    client_printf(core, user, "SONGS: %d\n", core->stats.songs);
    if(has_flag(core->status, STATUS_PLAYING) && 
       !has_flag(core->status, STATUS_PAUSED))
       client_printf(core, user, "PLAYED: %d\n", core->stats.playtime + 
		     (time(0) - core->stats.song_start));
    else
	client_printf(core, user, "PLAYED: %d\n", core->stats.playtime);
    client_printf(core, user, "LISTSIZE: %d\n", 
		  list_size(&core->active_playlist->songs));
    return(0);
}

int cmd_mute(core *core, client *user, char *param)
{
    if(core || param || user) {}
    log_printf(core, "cmd_mute() called");
    return(0);
}

int cmd_volume(core *core, client *user, char *param)
{
    if(core || param || user) {}
    log_printf(core, "cmd_volume() called");
    return(0);
}


int cmd_showplugins(core *core, client *user, char *param)
{
    int pos;
    list_node *plugin;
    if(core || param || user) {}
    if(!param || strcasecmp(param, "INPUT") == 0)
    {
	pos = 0;
	list_for_each(plugin, &core->input)
	{
	    client_printf(core, user, "PLUGIN INPUT %d: %s\n", pos++,
			  ((input_plugin*)plugin->data)->description);
	}
    }
    if(!param || strcasecmp(param, "OUPUT") == 0)
    {
	pos = 0;
	list_for_each(plugin, &core->output)
	{
	    client_printf(core, user, "PLUGIN OUTPUT %d: %s\n", pos++,
			  ((output_plugin*)plugin->data)->description);
	}
    }
    return(0);
}


int cmd_moduleload(core *core, client *user, char *param)
{
    if(core || param || user) {}
    log_printf(core, "cmd_moduleload() called");
    if(!param)
    {
	send_to_client(core, user, "ERROR: Missing filename.\n");
	return(-1);
    }
    switch(plugin_load(core, param))
    {
        case -1:
	    send_to_client(core, user, "ERROR: Opening module failed.\n");
	    break;
	    
        case 0:
	    client_printf(core, user, "Module %s loaded.\n", param);
	    break;
	    
        case 1:
	    send_to_client(core, user, "ERROR: Not a plugin.\n");
    }
    return(0);
}


int cmd_modulescandir(core *core, client *user, char *param)
{
    if(core || param || user) {}
    log_printf(core, "cmd_modulescandir() called");
    if(!param)
    {   
	send_to_client(core, user, "ERROR: Missing directory.\n");
	return(-1);
    }
    list_add(strdup(param), list_size(&core->modpaths) + 1, &core->modpaths);
    plugin_scan_dir(core, param);
    return(0);
}


int cmd_save(core *core, client *user, char *param)
{
    list_node *node;
    char *backup = calloc(1, strlen(core->configfile) + 5);
    FILE *fd;
    
    if(core || param || user) {}
    log_printf(core, "cmd_save() called");
    sprintf(backup, "%s.bak", core->configfile);
    if(rename(core->configfile, backup))
    {
	send_to_client(core, user, "ERROR: Unable to backup configfile.\n");
	return(-1);
    }
    
    fd = fopen(core->configfile, "w");
    
    if(!fd)
    {
	send_to_client(core, user, "ERROR: Unable to create a new file\n");
	return(-1);
    }
    
    fprintf(fd, "# Configuration file for Music daemon\n");
    fprintf(fd, "# Don't edit unless you know what you are doing\n\n");

    fprintf(fd, "set random=%s\n", 
	    has_flag(core->status, STATUS_RANDOM) ? "on" : "off");

    fprintf(fd, "set listen-port=%d\n", core->listen_port);

    fprintf(fd, "set unix-sock=%s\n", core->unixfile ? core->unixfile : "0");
    
    list_for_each(node, &core->input)
	fprintf(fd, "moduleload %s\n", ((input_plugin*)node->data)->filename);

    list_for_each(node, &core->output)
	fprintf(fd, "moduleload %s\n", ((output_plugin*)node->data)->filename);

    list_for_each(node, &core->modpaths)
	fprintf(fd, "modulescandir %s\n", (char*)node->data);

    fclose(fd);
    return(0);
}
    
