/* init.c -- Initialization for musicdaemon player
 *
 * Copyright (C) 2002-2003 Petri Lahtinen <pjl@timefade.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#include "config.h"

#include "main.h"
#include "network.h"
#include "init.h"
#include "parse.h"
#include "settings.h"

int init(core *core, int argc, char *argv[])
{
    playlist *playlist;

    /* Defaults and other initialization */
    list_init(&core->clients);
    list_init(&core->playlist);
    list_init(&core->input);
    list_init(&core->output);
    
    /* Disable as default, should be defined in config or cmdline */
    core->network = -1;
    core->unixdomain = -1;
    core->logfd = -1;

    core->seed = time(NULL);
    srand(core->seed);

    /* Create the default playlist */
    playlist = calloc(1, sizeof(struct playlist));
    playlist->position = 1;
    playlist->name = strdup("main");
    list_init(&playlist->songs);
    list_add(playlist, 0, &core->playlist);
    core->active_playlist = playlist;
    
    core->stats.uptime = time(0);
    core->stats.songs = 0;
    core->stats.playtime = 0;


    
    /* Star loading */
    
    if(parse_preload_cmdline(core, argc, argv) == -1)
	return(-1);
    
    core->configfile = get_configfile(argc, argv);
    
    if(!core->configfile)
    {
	fprintf(stderr, "No configfile found.\n");
    }
    else
    {
	printf("Using configuration: %s\n", core->configfile);
    
	if(read_config(core, core->configfile))
	    return(-1);
    }
    
    return(parse_postload_cmdline(core, argc, argv));
}


char *get_configfile(int argc, char *argv[])
{
    int i;
    char *file;
    char *home;
    
    struct stat buf;
    
    for(i = 1; i < argc; i++)
    {
	if(strcmp(argv[i], "-c") == 0)
	    return(strdup(argv[i+1]));
    }
    
    home = getenv("HOME");
    if(home)
    {
	file = calloc(1, strlen(home) + 20);
	sprintf(file, "%s/.musicd/config", home);
	if(stat(file, &buf) == 0)
	    return(file);
    }
    else
	fprintf(stderr, "Cannot determine home directory.\n");
    
    if(stat(CONFIG_DIR"/musicd.conf", &buf) == 0)
	return(strdup(CONFIG_DIR"/musicd.conf"));
    
    return(0);
}


int parse_postload_cmdline(core *core, int argc, char *argv[])
{
    int i;
    
    for(i = 0; i < argc; i++)
    {
	if(strcmp(argv[i], "-c") == 0 || strcmp(argv[i] ,"--config") == 0)
	    continue;

	else if(strcmp(argv[i], "-p") == 0 || strcmp(argv[i], "--port") == 0)
	{
	    i++;
	    if(i >= argc)
	    {
		fprintf(stderr, "Missing port number.\n");
		return(-1);
	    }
	    switch( set_listenport(core, argv[i]))
	    {
	        case -1:
		    fprintf(stderr, "Invalid port number.\n");
		    return(-1);
		    break;
	        case 0:
		    break;
	        default:
		    printf("BUG: set_listenport() returned an invalid value.\n");
		    break;
	    }
	}
    }
    return(0);
}


int parse_preload_cmdline(core *core, int argc, char *argv[])
{
    int i;
    
    if(core){}
    
    for(i = 0; i < argc; i++)
    {
	if(strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "--help") == 0)
	{
	    print_help();
	    return(-1);
	}
	if(strcmp(argv[i], "-v") == 0 || strcmp(argv[i], "--version") == 0)
	{
	    printf("Musicdaemon version %s\n", VERSION);
	    return(-1);
	}
    }
    return(0);
}


int read_config(core *core, char *filename)
{
    FILE *fd;
    char buf[MAX_MSG_SIZE + 1];
    
    fd = fopen(filename, "r");
    
    if(!fd)
    {
	fprintf(stderr, "Could not open configfile \"%s\" for reading.\n",
		filename);
	return(-1);
    }
    
    while(!feof(fd))
    {
	if(!fgets(buf, sizeof(buf), fd))
	    break;

	if(buf[0] == '#' || buf[0] == 0)
	    continue;

	if(!strchr(buf, '\n'))
	{
	    while(!strchr(buf, '\n'))
	    {
		*buf = 0;
		fgets(buf, sizeof(buf), fd);
	    }
	    fprintf(stderr, "Too long line in configfile (skipped).\n");
	    continue;
	}
	
	(*strchr(buf,'\n')) = 0;
	
	parse_line(core, NULL, buf);
    }
    
    fclose(fd);
    return(0);
}


void print_help()
{
    printf("Musicdaemon version %s.\n", VERSION);
    printf("Usage: musicd [option(s)]\n");
    printf("  -c, --config <file>       Configuration file to use.\n");
    printf("  -h, --help                Print this usefull help.\n");
    printf("  -p, --port <num>          Bind to port <num>.\n");
    printf("  -v, --version             Display musicd version.\n");
    return;
}
