/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 8 -*- */

/* network.c -- Networking code for musicdaemon player
 *
 * Copyright (C) 2002-2003 Petri Lahtinen <pjl@timefade.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netdb.h>
#include <netinet/in.h>

#include "network.h"
#include "parse.h"
#include "config.h"
#include "log.h"
#include "list.h"

#define UNIX_PATH_MAX	108


int listen_for_connections(int port)
{
    int sock;

#ifdef HAVE_IPV6

    struct sockaddr_in6 server6;
    memset(&server6, 0, sizeof(struct sockaddr_in6));
    sock = socket(PF_INET6, SOCK_STREAM, 0);
    server6.sin6_family = PF_INET6;
    server6.sin6_addr = in6addr_any;
    server6.sin6_port = htons(port);

    if(sock < 0)
    {
	if(errno == EINVAL || errno == EAFNOSUPPORT)
	{
#endif
	    struct sockaddr_in server4;
	    memset(&server4, 0, sizeof(struct sockaddr_in));
	    sock = socket(PF_INET, SOCK_STREAM, 0);
	    server4.sin_family = PF_INET;
	    server4.sin_addr.s_addr = INADDR_ANY;
	    server4.sin_port = htons(port);

	    if(sock < 0)
	    {
		perror("socket failed");
		return(-1);
	    }
	    
	    if(bind(sock, (struct sockaddr *) &server4, 
		    sizeof(struct sockaddr_in)))
	    {   
		perror("bind failed");
		return(-1);
	    }

#ifdef HAVE_IPV6

	}
	else
	{
	    perror("socket failed");
	    return(-1);
	}
    }
    else
    {
	if(bind(sock, (struct sockaddr *) &server6, 
		sizeof(struct sockaddr_in6)))
	{
	    perror("bind failed");
	    return(-1);
	}
    }

#endif

    if(listen(sock, 5))
    {
	perror("listen failed");
	return(-1);
    }
    
    return(sock);
}


int open_unix_socket(char *filename)
{
    int sock;
    struct sockaddr_un server;
    
    memset(&server, 0, sizeof(struct sockaddr_un));
    
    sock = socket(PF_UNIX, SOCK_STREAM, 0);
    
    server.sun_family = AF_UNIX;
    strncpy(server.sun_path, filename, UNIX_PATH_MAX-1);
    
    if(bind(sock, (struct sockaddr *) &server, sizeof(struct sockaddr_un)))
    {   
	perror("bind failed");
	return(-1);
    }
    
    if(listen(sock, 5))
    {   
	perror("listen failed");
	return(-1);
    }
    
    return(sock);
}


int load_fds(core *core, fd_set *read, fd_set *write)
{
    list_node *point;
    int n = 0;
    int s;

    FD_ZERO(read);
    FD_ZERO(write);

    if((s = core->unixdomain) >= 0)
    {
	FD_SET(s, read);
	if(s > n)
	    n = s;
    }

    if((s = core->network) >= 0)
    {
	FD_SET(s, read);
	if(s > n)
	    n = s;
    }

    list_for_each(point, &core->clients)
    {
	s = ((client *) point->data)->socket;
	if(((client *) point->data)->outbuf.start)
	{
	    FD_SET(s, write);
	}
	FD_SET(s, read);
	if(s > n)
	    n = s;
    }
    return(n);
}


int wait_input(core *core, struct timeval *timeout)
{
    int n;
    int num_fds;
    int ret = 0;
    fd_set read, write;

    n = load_fds(core, &read, &write);

    num_fds = select(n + 1, &read, &write, 0, timeout);

    if(num_fds > 0)
    {
	read_input(core, &read, num_fds);
	write_output(core, &write, num_fds);
	ret = 0;
    }
    else if(num_fds == -1)
    {
	if(errno != EINTR)
	{
	    perror("select failed");
	    ret = 1;
	}
    }

    FD_ZERO(&read);
    FD_ZERO(&write);

    return(ret);
}


int read_input(core *core, fd_set *fd, int num)
{
    list_node *point;
    list_node *next;

    if(core->network >= 0 && FD_ISSET(core->network, fd))
    {
	new_client(core, core->network);
	num--;
	if(!num)
	    return(0);
    }
    if(core->unixdomain >= 0 && FD_ISSET(core->unixdomain, fd))
    {
	new_client(core, core->unixdomain);
	num--;
	if(!num)
	    return(0);
    }

    for(point = core->clients.start; point; point = next)
    {
	if(!num)
	    break;
	next = point->next;
	if(FD_ISSET(((client *) point->data)->socket, fd))
	{
	    read_from_client(core, point->data);
	    num--;
	}
    }
    return(0);
}

int write_output(core *core, fd_set *fd, int num)
{
    int sent;
    list_node *point;
    client *user;

    list_for_each(point, &core->clients)
    {
	if(!num)
	    break;
	if(FD_ISSET(((client *) point->data)->socket, fd))
	{
	    user = ((client *) point->data);
	    sent = write(user->socket,
			 user->outbuf.start->data,
			 strlen(user->outbuf.start->data));
	    free(user->outbuf.start->data);
	    list_delete(0, &user->outbuf);
	    num--;
	}
    }
    return(0);
}


int new_client(core *core, int socket)
{
    client *new = calloc(1, sizeof(struct client));

    log_printf(core, "New connection!");

    new->socket = accept(socket, 0, 0);

    list_init(&new->outbuf);

    send_to_client(core, new, "Hello\n");

    list_add(new, 0, &core->clients);

    return(0);
}

int remove_client(core *core, client *user)
{
    int pos = 1;
    list_node *point;

    log_printf(core, "Client lost.");

    list_for_each(point, &core->clients)
    {
	if(point->data == user)
	{
	    close(user->socket);
	    while(user->outbuf.start)
	    {
		free(user->outbuf.start->data);
		list_delete(0, &user->outbuf);
	    }
	    free(user);
	    list_delete(pos, &core->clients);
	    break;
	}
	pos++;
    }

    return(0);
}


int read_from_client(core *core, client *user)
{
    int read_data;

    read_data = read(user->socket, user->buffer + user->bufferfill,
		     sizeof(user->buffer) - user->bufferfill - 1);
    if(read_data == 0)
    {
	remove_client(core, user);
	return(0);
    }
    if(read_data == -1)
    {
	if(errno != EINTR)
	{
	    remove_client(core, user);
	}
	return(0);
    }

    user->bufferfill += read_data;
    user->buffer[user->bufferfill] = 0;

    parse_client(core, user);

    return(0);
}


int send_to_client(core *core, client *user, char *message)
{
    char *msg;
    if(!user)
    {
	char *linefeed = strrchr(message, '\n');
	if(linefeed)
	    *linefeed = 0;
	log_printf(core, "Message for nobody: %s", message);
	return(1);
    }
    msg = strdup(message);
    if(!msg)
    {
	perror("strdup failed");
	exit(123);
    }
    if(core) {}
    list_add(msg, list_size(&user->outbuf) + 1, &user->outbuf);
    return(0);
}


int client_printf(core *core, client *user, char *message, ...)
{
    char buf[MAX_MSG_SIZE+1];
    va_list ap;
    
    va_start(ap, message);
    vsnprintf(buf, MAX_MSG_SIZE, message, ap);
    buf[MAX_MSG_SIZE] = 0;
    send_to_client(core, user ,buf);
    
    return(0);
}


int broadcast(core *core, char *message)
{
    list_node *point;
    list_for_each(point, &core->clients)
	send_to_client(core, point->data, message);

    return(0);
}

void network_close(core *core)
{
    if(core->network >= 0)
    {
	shutdown(core->network, SHUT_RDWR);
	close(core->network);
    }
    if(core->unixdomain >= 0)
    {
	shutdown(core->unixdomain, SHUT_RDWR);
	close(core->unixdomain);
	unlink(core->unixfile);
    }
}


int broad_printf(core *core, char *message, ...)
{
    char buf[MAX_MSG_SIZE+1];
    va_list ap;
    
    va_start(ap, message);
    vsnprintf(buf, MAX_MSG_SIZE, message, ap);
    buf[MAX_MSG_SIZE] = 0;
    broadcast(core, buf);
    
    return(0);
}
