/* oss.c -- OSS-plugin for musicdaemon player
 *
 * Copyright (C) 2002-2003 Petri Lahtinen <pjl@timefade.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <pthread.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>

#include "../../musicd.h"

static int dsp;
static char buffer[100000];
static int begin;
static int end;
static int pausing;
static int playing;
static pthread_t output_thread;
static pthread_mutex_t in_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t out_mutex = PTHREAD_MUTEX_INITIALIZER;

output_plugin *get_oplugin_info(void);
static void init(void);
static int open_audio(audio_format fmt, int rate, int nch);
static void write_audio(void *ptr, int length);
static void close_audio(void);
static void flush(int time);
static void do_pause(short paused);
static int buffer_free(void);
static int buffer_playing(void);
static int output_time(void);
static int written_time(void);
static void* play_loop(void*);

output_plugin meta_output_op =
{
    42,
    NULL,
    NULL,
    "OSS output plugin",
    init,
    NULL,
    NULL,
    open_audio,
    write_audio,
    close_audio,
    flush,
    do_pause,
    buffer_free,
    buffer_playing,
    output_time,
    written_time
};

output_plugin *get_oplugin_info()
{
    return(&meta_output_op);
}

static void init()
{
    begin = 0;
    end = 0;
}

static int open_audio(audio_format fmt, int rate, int nch)
{
    int format;

    switch(fmt)
    {
        case FMT_U8:
	    format = AFMT_U8;
	    break;
        case FMT_S8:
	    format = AFMT_S8;
	    break;
        case FMT_U16_LE:
	    format = AFMT_U16_LE;
	    break;
        case FMT_U16_BE:
	    format = AFMT_U16_BE;
	    break;
        case FMT_S16_LE:
	    format = AFMT_S16_LE;
	    break;
        case FMT_S16_BE:
	    format = AFMT_S16_BE;
	    break;
        default:
	    fprintf(stderr,"unknown audio format\n");
	    return(0);
    }
		  
    begin = 0;
    end = 0;

    dsp = open("/dev/dsp", O_WRONLY);
    if(dsp < 0)
    {
	perror("open failed");
	return(0);
    }
    if(ioctl(dsp, SNDCTL_DSP_RESET) == -1)
    {   
	perror("SNDCTL_DSP_RESET");
	return(0);
    }
    if(ioctl(dsp, SNDCTL_DSP_SETFMT, &format) == -1)
    {   
	perror("SNDCTL_DSP_SETFMT");
	return(0);
    }
    if(ioctl(dsp, SNDCTL_DSP_CHANNELS, &nch) == -1)
    {   
	perror("SNDCTL_DSP_CHANNELS");
	return(0);
    }
    if(ioctl(dsp, SNDCTL_DSP_SPEED, &rate) == -1)
    {   
	perror("SNDCTL_DSP_SPEED");
	return(0);
    }

    playing = 1;
    pthread_create(&output_thread, NULL, play_loop, NULL);
    
    return(1);
}

static void* play_loop(void* none)
{
    int written;
    int data;

    if(none){}
    
    while(playing)
    {
	while(pausing)
	    usleep(10000);
	pthread_mutex_lock(&out_mutex);
	data = end-begin;
	if(data < 0) data += sizeof(buffer);
	if(data > 1024) data = 1024;
	if(data)
	{
	    if(data < (int)sizeof(buffer)-begin)
	    {
		written = write(dsp, &buffer[begin], data);
		if(written == -1)
		{
		    perror("write failed");
		    pthread_mutex_unlock(&out_mutex);
		    continue;
		}
		begin += written;
		if(begin >= (int)sizeof(buffer)) begin -= sizeof(buffer);
	    }
	    else
	    {
		written = write(dsp, &buffer[begin], sizeof(buffer)-begin );
		if(written == -1)
		{
		    perror("write failed");
		    pthread_mutex_unlock(&out_mutex);
		    continue;
		}
		begin += written;
		if(begin >= (int)sizeof(buffer)) begin -= sizeof(buffer);
	    }
	    pthread_mutex_unlock(&out_mutex);
	}
	else
	{
	    pthread_mutex_unlock(&out_mutex);
	    usleep(100000);
	}
    }
    playing = 0;
    return(0);
}

static void write_audio(void *ptr, int length)
{
    char *data = ptr;
    int free_buffer = begin - end;

    pthread_mutex_lock(&in_mutex);

    free_buffer = begin - end;
    if(free_buffer < 0) free_buffer += sizeof(buffer);
    
    if(length > (int)sizeof(buffer) - end)
    {
	memcpy(&buffer[end], data, sizeof(buffer)-end);
	memcpy(buffer, &data[sizeof(buffer)-end], length - (sizeof(buffer)-end));
	end = length - (sizeof(buffer)-end);
    }
    else
    {
	memcpy(&buffer[end], data, length);
	end += length;
    }
    pthread_mutex_unlock(&in_mutex);
}

static void close_audio()
{
    if(playing)
    {
	playing = 0;
	pthread_join(output_thread, NULL);
    }
    if(dsp)
    {
	close(dsp);
	dsp = 0;
    }
}

static void do_pause(short paused)
{
    pausing = paused;
    return;
}

static int buffer_free()
{
    int free_buffer = begin - end;
        
    if(free_buffer <= 0) free_buffer += sizeof(buffer);

    return(free_buffer-1);
}

static int buffer_playing()
{
    return(1);
}

static int output_time()
{
    return(1);
}

static int written_time()
{
    return(1);
}

static void flush(int time)
{
    if(time){}

    pthread_mutex_lock(&in_mutex);
    pthread_mutex_lock(&out_mutex);
    end = begin = 0;
    pthread_mutex_unlock(&in_mutex);
    pthread_mutex_unlock(&out_mutex);
}
