/* solaris.c -- Solaris-plugin for musicdaemon player
 *
 * Copyright (C) 2002-2003 Petri Lahtinen <pjl@timefade.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


#include <pthread.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/audio.h>
#include <sys/audioio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <stropts.h>

#include "../../musicd.h"

#include "config.h"

static int dsp;
static char buffer[100000];
static int begin;
static int end;
static int pausing;
static int playing;
static int swap_endian;
static pthread_t output_thread;
static pthread_mutex_t in_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t out_mutex = PTHREAD_MUTEX_INITIALIZER;

output_plugin *get_oplugin_info(void);
static void init(void);
static int open_audio(audio_format fmt, int rate, int nch);
static void write_audio(void *ptr, int length);
static void close_audio(void);
static void flush(int time);
static void do_pause(short paused);
static int buffer_free(void);
static int buffer_playing(void);
static int output_time(void);
static int written_time(void);
static void* play_loop(void*);
static void swap_buffer_endian(char *buf, int len);

output_plugin meta_output_op =
{
    42,
    NULL,
    NULL,
    "Solaris output plugin",
    init,
    NULL,
    NULL,
    open_audio,
    write_audio,
    close_audio,
    flush,
    do_pause,
    buffer_free,
    buffer_playing,
    output_time,
    written_time
};

output_plugin *get_oplugin_info()
{
    return(&meta_output_op);
}

static void init()
{
    begin = 0;
    end = 0;
}

static int open_audio(audio_format fmt, int rate, int nch)
{
    audio_info_t info;

    begin = 0;
    end = 0;

#ifdef WORDS_BIGENDIAN
    if(fmt == FMT_S16_LE || fmt == FMT_U16_LE)
	swap_endian = 1;
    else
	swap_endian = 0;
#else
    if(fmt == FMT_S16_BE || fmt == FMT_U16_BE)
	swap_endian = 1;
    else
	swap_endian = 0;
#endif

    
    dsp = open("/dev/audio", O_WRONLY);
    if(dsp < 0)
    {
	perror("open failed");
	return(0);
    }
    if(ioctl(dsp, AUDIO_GETINFO, &info) == -1)
    {   
	perror("AUDIO_GETINFO");
	return(0);
    }

    info.play.sample_rate = rate;
    info.play.channels = nch;

    if(fmt == FMT_U8 || fmt == FMT_S8)
	info.play.precision = 8;
    else
	info.play.precision = 16;

    info.play.encoding = AUDIO_ENCODING_LINEAR;

    if(ioctl(dsp, AUDIO_SETINFO, &info) == -1)
    {   
	perror("AUDIO_SETINFO");
	return(0);
    }
    playing = 1;
    pthread_create(&output_thread, NULL, play_loop, NULL);
    
    return(1);
}

static void* play_loop(void* none)
{
    int written;
    int data;

    if(none){}
    
    while(playing)
    {
	while(pausing)
	    usleep(10000);
	pthread_mutex_lock(&out_mutex);
	data = end-begin;
	if(data < 0) data += sizeof(buffer);
	if(data > 15)
	{
	    if(data < (int)sizeof(buffer)-begin)
	    {
	        written = write(dsp, &buffer[begin], data&(~15));
		if(written == -1)
		{
		    perror("write failed");
		    pthread_mutex_unlock(&out_mutex);
		    continue;
		}

		begin += written;
		if(begin >= (int)sizeof(buffer)) begin -= sizeof(buffer);
	    }
	    else
	    {
		written = write(dsp, &buffer[begin], sizeof(buffer)-begin );
		if(written == -1)
		{
		    perror("write failed");
		    pthread_mutex_unlock(&out_mutex);
		    continue;
		}

		begin += written;
		if(begin >= (int)sizeof(buffer)) begin -= sizeof(buffer);
	    }
	    pthread_mutex_unlock(&out_mutex);
	}
	else
	{
	    pthread_mutex_unlock(&out_mutex);
	    usleep(100000);
	}
    }
    playing = 0;
    return(0);
}

static void write_audio(void *ptr, int length)
{
    char *data = ptr;
    int free_buffer = begin - end;
    pthread_mutex_lock(&in_mutex);
    if(swap_endian)
	swap_buffer_endian(data, length);
    free_buffer = begin - end;
    if(free_buffer < 0) free_buffer += sizeof(buffer);
    
    if(length > (int)sizeof(buffer) - end)
    {
	memcpy(&buffer[end], data, sizeof(buffer)-end);
	memcpy(buffer, &data[sizeof(buffer)-end], length - (sizeof(buffer)-end));
	end = length - (sizeof(buffer)-end);
    }
    else
    {
	memcpy(&buffer[end], data, length);
	end += length;
    }
    pthread_mutex_unlock(&in_mutex);
}

static void close_audio()
{
    if(playing)
    {
	playing = 0;
	pthread_join(output_thread, NULL);
    }
    if(dsp)
    {
	close(dsp);
	dsp = 0;
    }
}

static void do_pause(short paused)
{
    pausing = paused;
    return;
}

static int buffer_free()
{
    int free_buffer = begin - end;
        
    if(free_buffer <= 0) free_buffer += sizeof(buffer);

    return((free_buffer-1)&(~15));
}

static int buffer_playing()
{
    return(1);
}

static int output_time()
{
    return(1);
}

static int written_time()
{
    return(1);
}

static void flush(int time)
{
    if(time){}
    pthread_mutex_lock(&in_mutex);
    pthread_mutex_lock(&out_mutex);
    end = begin = 0;
    pthread_mutex_unlock(&in_mutex);
    pthread_mutex_unlock(&out_mutex);
}

static void swap_buffer_endian(char *buf, int len)
{
    int i;
    char tmp;
    for(i = 0; i < len; i++)
    {
	tmp = buf[2*i];
	buf[2*1] = buf[2*i+1];
	buf[2*i+1] = tmp;
    }
    return;
}
