/* settings.c -- Settings for musicdaemon player
 *
 * Copyright (C) 2002-2003 Petri Lahtinen <pjl@timefade.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/socket.h>
#include <unistd.h>

#include "main.h"
#include "settings.h"
#include "network.h"


/* Yes, it is a global variable :( */

setting settings[] = { 
    {"LISTEN-PORT", set_listenport, print_listenport},
    {"LOOP",        set_loop,       print_loop},
    {"PLAYLIST",    set_playlist,   print_playlist},
    {"RANDOM",      set_randomplay, print_randomplay},
    {"UNIX-SOCK",   set_unixsock,   print_unixsock},
    {NULL,          NULL,           NULL}
};


int set(core *core, char *variable, char *value)
{
    int tmp;
    
    for(tmp = 0; settings[tmp].variable; tmp++)
    {
	if(strcasecmp(variable, settings[tmp].variable) == 0)
	{
	    return( (settings[tmp].set_val)(core, value) );
	}
    }
    return(1);
}

int setting_print(core *core, client *user, char *variable)
{
    int tmp;
    
    for(tmp = 0; settings[tmp].variable; tmp++)
    {
	if(strcasecmp(variable, settings[tmp].variable) == 0)
	{
	    return( (settings[tmp].print_val)(core, user) );
	}
    }
    return(1);    
}


int settings_print_all(core *core, client *user)
{
    int tmp;
    
    for(tmp = 0; settings[tmp].variable; tmp++)
    {
	(settings[tmp].print_val)(core, user);
    }
    return(0);
}


/* Check string for bool value setting (yes/no, on/off, 1/0,
 * true/false). We're not too picky about the actual format. Return bool value
 * or -1 for failure.
 */

int strtobool(char *str)
{
    if(!str)
	return(-1);
    
    if( strcasecmp(str, "yes") == 0 || strcasecmp(str, "on") == 0 || 
	strcasecmp(str, "true") == 0 || strcasecmp(str, "1") == 0 )
    {
	return(1);
    }

    if( strcasecmp(str, "no") == 0 || strcasecmp(str, "off") == 0 || 
	strcasecmp(str, "false") == 0 || strcasecmp(str, "0") == 0 )
    {
	return(0);
    }
    return(-1);
}

int set_randomplay(core *core, char *value)
{
    int val = strtobool(value);
    
    if(val == -1)
	return(-1);
    
    if(val)
	set_flag(core->status, STATUS_RANDOM);
    else
	unset_flag(core->status, STATUS_RANDOM);
    
    return(0);
}

int print_randomplay(core *core, client *user)
{
    int val = has_flag(core->status, STATUS_RANDOM) ? 1 : 0;
    client_printf(core, user, "VALUE: RANDOM=%d\n", val);
    return(0);
}

int set_playlist(core *core, char *value)
{
    list_node *plist;
    playlist *this;
    
    list_for_each(plist, &core->playlist)
    {
	this = (playlist *)plist->data;
	if( strcmp(this->name, value) == 0 )
	{
	    core->active_playlist = this;
	    return(0);
	}
    }
    
    this = calloc(1, sizeof(struct playlist));
    this->position = 1;
    this->name = strdup(value);
    list_init(&this->songs);
    list_add(this, 0, &core->playlist);
    core->active_playlist = this;
    
    return(0);
}

int print_playlist(core *core, client *user)
{
    client_printf(core, user, "VALUE: PLAYLIST=%s\n", 
		  core->active_playlist->name);
    return(0);
}
    
int set_listenport(core *core, char *value)
{
    int port = atoi(value);
    int sock;
    
    if(strcmp(value, "0") == 0)
    {
	printf("Disabling network.\n");
	if(core->network)
	{
	    shutdown(core->network, SHUT_RDWR);
	    close(core->network);
	}
	core->listen_port = 0;
	core->network = -1;
	return(0);
    }
    
    if(!(port < 65536 && port > 0))
	return(-1);
    
    sock = listen_for_connections(port);
    
    if(sock >= 0)
    {
	core->listen_port = port;
	printf("Binding to port %d.\n", port);
    
	if(core->network)
	{
	    shutdown(core->network, SHUT_RDWR);
	    close(core->network);
	}

	core->network = sock;
    }
    
    return(0);
}

int print_listenport(core *core, client *user)
{
    if(core->listen_port)
	client_printf(core, user, "VALUE: LISTEN-PORT=%d\n", 
		      core->listen_port);
    else
	client_printf(core, user, "VALUE: LISTEN-PORT=<disabled>\n");
    return(0);
}
    
int set_loop(core *core, char *value)
{
    int val = strtobool(value);
    
    if(val == -1)
	return(-1);
    
    if(val)
	set_flag(core->status, STATUS_LOOP);
    else
	unset_flag(core->status, STATUS_LOOP);
    
    return(0);
}

int print_loop(core *core, client *user)
{
    int val = has_flag(core->status, STATUS_LOOP) ? 1 : 0;
    client_printf(core, user, "VALUE: LOOP=%d\n", val);
    return(0);
}

int set_unixsock(core *core, char *value)
{
    int sock;

    if(strcmp(value, "0") == 0)
    {
	printf("Disabling unix-socket.\n");
	if(core->unixdomain)
	{
	    shutdown(core->unixdomain, SHUT_RDWR);
	    close(core->unixdomain);
	}
	free(core->unixfile);
	core->unixfile = 0;
	core->unixdomain = -1;
	return(0);
    }
    
    sock = open_unix_socket(value);
    if(sock >= 0)
    {
	printf("Listening in unix-socket %s\n", value);
    
	if(core->unixdomain)
	{   
	    shutdown(core->unixdomain, SHUT_RDWR);
	    close(core->unixdomain);
	    unlink(core->unixfile);
	}

	free(core->unixfile);
	core->unixfile = strdup(value);
	core->unixdomain = sock;
	if(core->unixdomain == -1)
	{
	    free(core->unixfile);
	    core->unixfile = 0;
	}
    }
       
    return(0);
}

int print_unixsock(core *core, client *user)
{
    if(core->unixfile)
	client_printf(core, user, "VALUE: UNIX-SOCK=%s\n", core->unixfile);
    else
	client_printf(core, user, "VALUE: UNIX-SOCK=<disabled>\n");
    return(0);
}
